/**
 * OW2 Util
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.util.jmx;

import java.util.HashSet;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanOperationInfo;
import javax.management.MBeanParameterInfo;
import javax.management.ReflectionException;
import javax.management.RuntimeOperationsException;
import javax.management.modelmbean.ModelMBeanInfo;

import org.apache.commons.modeler.BaseModelMBean;
import org.apache.commons.modeler.modules.ModelerSource;

/**
 * Extends Commons Modeler BaseModelMBean class to allow adding attributes and operations at runtime.
 * @author missonng
 */
public class BaseModelMBeanExt extends BaseModelMBean {
    /**
     * The dynamically added attributes.
     */
    private HashSet<AbstractMBeanAttribute> attributes = new HashSet<AbstractMBeanAttribute>();

    /**
     * The dynamically added operations.
     */
    private HashSet<AbstractMBeanOperation> operations = new HashSet<AbstractMBeanOperation>();

    /**
     * Construct a ModelMBean with default ModelMBeanInfo information.
     * @throws RuntimeOperationsException If the initializer of an object throws an exception.
     * @throws MBeanException If an IllegalArgumentException occurs.
     */
    public BaseModelMBeanExt() throws RuntimeOperationsException, MBeanException  {
        super();
    }

    /**
     * Construct a ModelMBean associated with the specified ModelMBeanInfo information.
     * @param info ModelMBeanInfo for this MBean.
     * @throws RuntimeOperationsException If the initializer of an object throws an exception.
     * @throws MBeanException If an IllegalArgumentException occurs.
     */
    public BaseModelMBeanExt(final ModelMBeanInfo info) throws RuntimeOperationsException, MBeanException  {
        super(info);
    }

    /**
     * Construct a ModelMBean of a specified type.
     * The type can be a class name or the key used in one of the descriptors.
     * If no descriptor is available, we'll first try to locate one in the same package with the class, then use introspection.
     * The mbean resource will be created.
     * @param type Class name or the type key used in the descriptor.
     * @throws RuntimeOperationsException If the initializer of an object throws an exception.
     * @throws MBeanException If an IllegalArgumentException occurs.
     */
    public BaseModelMBeanExt(final String type) throws RuntimeOperationsException, MBeanException  {
        super(type);
    }

    /**
     * Construct a ModelMBean of a specified type.
     * The type can be a class name or the key used in one of the descriptors.
     * If no descriptor is available, we'll first try to locate one in the same package with the class, then use introspection.
     * The mbean resource will be created.
     * @param type Class name or the type key used in the descriptor.
     * @param source ???
     * @throws RuntimeOperationsException If the initializer of an object throws an exception.
     * @throws MBeanException If an IllegalArgumentException occurs.
     */
    public BaseModelMBeanExt(final String type, final ModelerSource source) throws RuntimeOperationsException, MBeanException  {
        super(type, source);
    }

    /**
     * Return the MBeanInfo object for this MBean.
     * @return The MBeanInfo object for this MBean.
     */
    @Override
    public synchronized MBeanInfo getMBeanInfo() {
        //Generate the attribute info from dynamically added attributes and those define statically by Commons Modeler.
        MBeanAttributeInfo[] attributeInfo1 = new MBeanAttributeInfo[0];
        if (this.attributes != null) {
            int i = 0;
            attributeInfo1 = new MBeanAttributeInfo[this.attributes.size()];
            for (AbstractMBeanAttribute attribute : this.attributes) {
                attributeInfo1[i++] = attribute.getMBeanAttributeInfo();
            }
        }
        MBeanAttributeInfo[] attributeInfo2 = super.getMBeanInfo().getAttributes();
        MBeanAttributeInfo[] attributeInfo = new MBeanAttributeInfo[attributeInfo1.length + attributeInfo2.length];
        System.arraycopy(attributeInfo1, 0, attributeInfo, 0, attributeInfo1.length);
        System.arraycopy(attributeInfo2, 0, attributeInfo, attributeInfo1.length, attributeInfo2.length);

        //Generate the operation info from dynamically added operations and those define statically by Commons Modeler.
        MBeanOperationInfo[] operationInfo1 = new MBeanOperationInfo[0];
        if (this.operations != null) {
            int i = 0;
            operationInfo1 = new MBeanOperationInfo[this.operations.size()];
            for (AbstractMBeanOperation operation : this.operations) {
                operationInfo1[i++] = operation.getMBeanOperationInfo();
            }
        }
        MBeanOperationInfo[] operationInfo2 = super.getMBeanInfo().getOperations();
        MBeanOperationInfo[] operationInfo = new MBeanOperationInfo[operationInfo1.length + operationInfo2.length];
        System.arraycopy(operationInfo1, 0, operationInfo, 0, operationInfo1.length);
        System.arraycopy(operationInfo2, 0, operationInfo, operationInfo1.length, operationInfo2.length);

        //Return the generated MBeanInfo.
        return new MBeanInfo(
                super.getMBeanInfo().getClassName(),
                super.getMBeanInfo().getDescription(),
                attributeInfo,
                super.getMBeanInfo().getConstructors(),
                operationInfo,
                super.getMBeanInfo().getNotifications());
    }

    /**
     * Add an attribute.
     * @param attribute The attribute to add.
     */
    public synchronized void addAttribute(final AbstractMBeanAttribute attribute) {
        this.attributes.add(attribute);
    }

    /**
     * Remove an attribute.
     * @param attribute The attribute to remove.
     */
    public synchronized void removeAttribute(final AbstractMBeanAttribute attribute) {
        this.attributes.remove(attribute);
    }

    /**
     * Obtain and return the values of several attributes of this MBean.
     * @param names Names of the requested attributes.
     * @return The values of several attributes of this MBean.
     */
    @Override
    public synchronized AttributeList getAttributes(final String[] names) {
        AttributeList attributeList = new AttributeList();
        for (String name : names) {
            try {
                attributeList.add(getAttribute(name));
            } catch (Exception ex) {
                //nothing to do.
                continue;
            }
        }

        return attributeList;
    }

    /**
     * Obtain and return the value of a specific attribute of this MBean.
     * @param name Name of the requested attribute.
     * @return The value of a specific attribute of this MBean.
     * @throws AttributeNotFoundException If this attribute is not supported by this MBean.
     * @throws MBeanException If the initializer of an object throws an exception.
     * @throws ReflectionException If a Java reflection exception occurs when invoking the getter.
     */
    @Override
    public synchronized Object getAttribute(final String name)
            throws AttributeNotFoundException, MBeanException, ReflectionException {
        try {
            //Try to get the value from dynamic attributes.
            AbstractMBeanAttribute iattribute = null;
            for (AbstractMBeanAttribute temp : this.attributes) {
                if (temp.getMBeanAttributeInfo().getName().equals(name)) {
                    iattribute = temp;
                    break;
                }
            }

            return iattribute.getValue();
        } catch (Exception ex) {
            //Invoke the super class method if an exception occur..
            return super.getAttribute(name);
        }
    }

    /**
     * Set the values of several attributes of this MBean.
     * @param attributes The names and values to be set
     * @return The list of attributes that were set and their new values.
     */
    @Override
    public synchronized AttributeList setAttributes(final AttributeList attributes) {
        AttributeList attributeList = new AttributeList();
        for (Object object : attributes) {
            try {
                setAttribute((Attribute) object);
                attributeList.add((Attribute) object);
            } catch (Exception ex) {
                //nothing to do.
                continue;
            }
        }

        return attributeList;
    }

    /**
     * Set the value of a specific attribute of this MBean.
     * @param attribute The identification of the attribute to be set and the new value.
     * @throws AttributeNotFoundException If this attribute is not supported by this MBean.
     * @throws MBeanException If the initializer of an object throws an exception.
     * @throws ReflectionException If a Java reflection exception occurs when invoking the getter.
     */
    @Override
    public synchronized void setAttribute(final Attribute attribute)
            throws AttributeNotFoundException, MBeanException, ReflectionException {
        try {
            AbstractMBeanAttribute iattribute = null;
            for (AbstractMBeanAttribute temp : this.attributes) {
                if (temp.getMBeanAttributeInfo().getName().equals(attribute.getName())) {
                    iattribute = temp;
                    break;
                }
            }

            iattribute.setValue(attribute.getValue());
        } catch (Exception ex) {
            super.setAttribute(attribute);
        }
    }

    /**
     * Add an operation.
     * @param operation The operation to add.
     */
    public synchronized void addOperation(final AbstractMBeanOperation operation) {
        this.operations.add(operation);
    }

    /**
     * Remove an operation.
     * @param operation The operation to remove.
     */
    public synchronized void removeOperation(final AbstractMBeanOperation operation) {
        this.operations.remove(operation);
    }

    /**
     * Invoke a particular method on this MBean, and return any returned value.
     * @param name Name of the operation to be invoked.
     * @param params Array containing the method parameters of this operation.
     * @param signature Array containing the class names representing the signature of this operation.
     * @return The value returned by the operation.
     * @throws MBeanException If the initializer of an object throws an exception.
     * @throws ReflectionException If a Java reflection exception occurs when invoking a method.
     */
    @Override
    public synchronized Object invoke(final String name, final Object[] params, final String[] signature)
            throws MBeanException, ReflectionException {
        try {
            AbstractMBeanOperation ioperation = null;
            for (AbstractMBeanOperation temp : this.operations) {
                MBeanParameterInfo[] tempSignature = temp.getMBeanOperationInfo().getSignature();
                if (signature.length == tempSignature.length) {
                    boolean ok = true;
                    for (int i=0;ok&&i<tempSignature.length;i++) {
                        if (!tempSignature[i].getType().equals(signature[i])) {
                            try {
                                ok = Class.forName(tempSignature[i].getType()).equals(Class.forName(signature[i]));
                            } catch (ClassNotFoundException e) {
                                ok = false;
                            }
                        }
                    }
                    if (ok) {
                        ioperation = temp;
                        break;
                    }
                }
            }

            return ioperation.invoke(params);
        } catch (Exception ex) {
            return super.invoke(name, params, signature);
        }
    }
}
